"""
nydus.db.base
~~~~~~~~~~~~~

:copyright: (c) 2011 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

from collections import defaultdict
from nydus.db.routers import BaseRouter
from nydus.utils import import_string, ThreadPool


def create_cluster(settings):
    """
    redis = create_cluster({
        'engine': 'nydus.db.backends.redis.Redis',
        'router': 'nydus.db.routers.redis.PartitionRouter',
        'hosts': {
            0: {'db': 0},
            1: {'db': 1},
            2: {'db': 2},
        }
    })
    """
    # Pull in our client
    if isinstance(settings['engine'], basestring):
        conn = import_string(settings['engine'])
    else:
        conn = settings['engine']

    # Pull in our router
    router = settings.get('router')
    if isinstance(router, basestring):
        router = import_string(router)()
    elif router:
        router = router()
    else:
        router = BaseRouter()

    # Build the connection cluster
    return Cluster(
        router=router,
        hosts=dict(
            (conn_number, conn(num=conn_number, **host_settings))
            for conn_number, host_settings
            in settings['hosts'].iteritems()
        ),
    )


class Cluster(object):
    """
    Holds a cluster of connections.
    """

    def __init__(self, hosts, router=None, max_connection_retries=20):
        self.hosts = hosts
        self.router = router
        self.max_connection_retries = max_connection_retries

    def __len__(self):
        return len(self.hosts)

    def __getitem__(self, name):
        return self.hosts[name]

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return self.__dict__[attr]
        else:
            return CallProxy(self, attr)

    def __iter__(self):
        for name in self.hosts.iterkeys():
            yield name

    def _execute(self, attr, args, kwargs):
        db_nums = self._db_nums_for(*args, **kwargs)

        if self.router and len(db_nums) is 1 and self.router.retryable:
            # The router supports retryable commands, so we want to run a
            # separate algorithm for how we get connections to run commands on
            # and then possibly retry
            return self._retryable_execute(db_nums, attr, *args, **kwargs)
        else:
            connections = self._connections_for(*args, **kwargs)
            results = [getattr(conn, attr)(*args, **kwargs) for conn in connections]

            # If we only had one db to query, we simply return that res
            if len(results) == 1:
                return results[0]
            else:
                return results

    def disconnect(self):
        """Disconnects all connections in cluster"""
        for connection in self.hosts.itervalues():
            connection.disconnect()

    def get_conn(self, *args, **kwargs):
        """
        Returns a connection object from the router given ``args``.

        Useful in cases where a connection cannot be automatically determined
        during all steps of the process. An example of this would be
        Redis pipelines.
        """
        connections = self._connections_for(*args, **kwargs)

        if len(connections) is 1:
            return connections[0]
        else:
            return connections

    def map(self, workers=None):
        return DistributedContextManager(self, workers)

    def _retryable_execute(self, db_nums, attr, *args, **kwargs):
        retries = 0

        while retries <= self.max_connection_retries:
            if len(db_nums) > 1:
                raise Exception('Retryable router returned multiple DBs')
            else:
                connection = self[db_nums[0]]

            try:
                return getattr(connection, attr)(*args, **kwargs)
            except tuple(connection.retryable_exceptions):
                # We had a failure, so get a new db_num and try again, noting
                # the DB number that just failed, so the backend can mark it as
                # down
                db_nums = self._db_nums_for(retry_for=db_nums[0], *args, **kwargs)
                retries += 1
        else:
            raise Exception('Maximum amount of connection retries exceeded')

    def _db_nums_for(self, *args, **kwargs):
        if self.router:
            return self.router.get_db(self, 'get_conn', *args, **kwargs)
        else:
            return range(len(self))

    def _connections_for(self, *args, **kwargs):
        return [self[n] for n in self._db_nums_for(*args, **kwargs)]


class CallProxy(object):
    """
    Handles routing function calls to the proper connection.
    """
    def __init__(self, cluster, attr):
        self._cluster = cluster
        self._attr = attr

    def __call__(self, *args, **kwargs):
        return self._cluster._execute(self._attr, args, kwargs)


class EventualCommand(object):
    _attr = None
    _wrapped = None
    _evaled = False

    # introspection support:
    __members__ = property(lambda self: self.__dir__())

    def __init__(self, attr):
        self._attr = attr
        self._wrapped = None
        self._evaled = False
        self._args = []
        self._kwargs = {}
        self._ident = None

    def __call__(self, *args, **kwargs):
        self._args = args
        self._kwargs = kwargs
        self._ident = ':'.join(map(str, [id(self._attr), id(self._args), id(self._kwargs)]))
        return self

    def __repr__(self):
        if self._evaled:
            return repr(self._wrapped)
        return u'<EventualCommand: %s args=%s kwargs=%s>' % (self._attr, self._args, self._kwargs)

    def __getattr__(self, name):
        if name in ('_attr', '_wrapped', '_evaled', '_args', '_kwargs', '_ident'):
            return getattr(self, name)
        return getattr(self._wrapped, name)

    def __setattr__(self, name, value):
        if name in ('_attr', '_wrapped', '_evaled', '_args', '_kwargs', '_ident'):
            return object.__setattr__(self, name, value)
        return setattr(self._wrapped, name, value)

    def __delattr__(self, name):
        if name in ('_attr', '_wrapped', '_evaled', '_args', '_kwargs', '_ident'):
            raise TypeError("can't delete %s." % name)
        delattr(self._wrapped, name)

    def __deepcopy__(self, memo):
        # Changed to use deepcopy from copycompat, instead of copy
        # For Python 2.4.
        from django.utils.copycompat import deepcopy
        return deepcopy(self._wrapped, memo)

    # Need to pretend to be the wrapped class, for the sake of objects that care
    # about this (especially in equality tests)
    def __get_class(self):
        return self._wrapped.__class__
    __class__ = property(__get_class)

    def _set_value(self, value):
        self._wrapped = value
        self._evaled = True

    def _execute(self, conn):
        return getattr(conn, self._attr)(*self._args, **self._kwargs)

    def __dict__(self):
        try:
            return self._current_object.__dict__
        except RuntimeError:
            return AttributeError('__dict__')
    __dict__ = property(__dict__)

    def __setitem__(self, key, value):
        self._wrapped[key] = value

    def __delitem__(self, key):
        del self._wrapped[key]

    def __setslice__(self, i, j, seq):
        self._wrapped[i:j] = seq

    def __delslice__(self, i, j):
        del self._wrapped[i:j]

    __delattr__ = lambda x, n: delattr(x._wrapped, n)
    __lt__ = lambda x, o: x._wrapped < o
    __le__ = lambda x, o: x._wrapped <= o
    __eq__ = lambda x, o: x._wrapped == o
    __ne__ = lambda x, o: x._wrapped != o
    __gt__ = lambda x, o: x._wrapped > o
    __ge__ = lambda x, o: x._wrapped >= o
    __cmp__ = lambda x, o: cmp(x._wrapped, o)
    __hash__ = lambda x: hash(x._wrapped)
    # attributes are currently not callable
    # __call__ = lambda x, *a, **kw: x._wrapped(*a, **kw)
    __nonzero__ = lambda x: bool(x._wrapped)
    __len__ = lambda x: len(x._wrapped)
    __getitem__ = lambda x, i: x._wrapped[i]
    __iter__ = lambda x: iter(x._wrapped)
    __contains__ = lambda x, i: i in x._wrapped
    __getslice__ = lambda x, i, j: x._wrapped[i:j]
    __add__ = lambda x, o: x._wrapped + o
    __sub__ = lambda x, o: x._wrapped - o
    __mul__ = lambda x, o: x._wrapped * o
    __floordiv__ = lambda x, o: x._wrapped // o
    __mod__ = lambda x, o: x._wrapped % o
    __divmod__ = lambda x, o: x._wrapped.__divmod__(o)
    __pow__ = lambda x, o: x._wrapped ** o
    __lshift__ = lambda x, o: x._wrapped << o
    __rshift__ = lambda x, o: x._wrapped >> o
    __and__ = lambda x, o: x._wrapped & o
    __xor__ = lambda x, o: x._wrapped ^ o
    __or__ = lambda x, o: x._wrapped | o
    __div__ = lambda x, o: x._wrapped.__div__(o)
    __truediv__ = lambda x, o: x._wrapped.__truediv__(o)
    __neg__ = lambda x: -(x._wrapped)
    __pos__ = lambda x: +(x._wrapped)
    __abs__ = lambda x: abs(x._wrapped)
    __invert__ = lambda x: ~(x._wrapped)
    __complex__ = lambda x: complex(x._wrapped)
    __int__ = lambda x: int(x._wrapped)
    __long__ = lambda x: long(x._wrapped)
    __float__ = lambda x: float(x._wrapped)
    __str__ = lambda x: str(x._wrapped)
    __unicode__ = lambda x: unicode(x._wrapped)
    __oct__ = lambda x: oct(x._wrapped)
    __hex__ = lambda x: hex(x._wrapped)
    __index__ = lambda x: x._wrapped.__index__()
    __coerce__ = lambda x, o: x.__coerce__(x, o)
    __enter__ = lambda x: x.__enter__()
    __exit__ = lambda x, *a, **kw: x.__exit__(*a, **kw)


class DistributedConnection(object):
    def __init__(self, cluster, workers=None):
        self._cluster = cluster
        self._workers = min(workers or len(cluster), 16)
        self._commands = []
        self._complete = False

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return self.__dict__[attr]
        else:
            command = EventualCommand(attr)
            self._commands.append(command)
            return command

    def _execute(self):
        num_commands = len(self._commands)
        if num_commands == 0:
            self._commands = []
            return

        command_map = {}
        pipelined = all(self._cluster[n].supports_pipelines for n in self._cluster)
        pending_commands = defaultdict(list)

        # used in pipelining
        if pipelined:
            pipe_command_map = defaultdict(list)

            pipes = dict()  # db -> pipeline

        # build up a list of pending commands and their routing information
        for command in self._commands:
            cmd_ident = command._ident

            command_map[cmd_ident] = command

            if self._cluster.router:
                db_nums = self._cluster.router.get_db(self._cluster, command._attr, *command._args, **command._kwargs)
            else:
                db_nums = range(len(self._cluster))

            # The number of commands is based on the total number of executable commands
            num_commands += len(db_nums)

            # Don't bother with the pooling if we only need to do one operation on a single machine
            if num_commands == 1:
                self._commands = [command._execute(self._cluster[n]) for n in n]
                return

            # update the pipelined dbs
            for db_num in db_nums:
                # map the ident to a db
                if pipelined:
                    pipe_command_map[db_num].append(cmd_ident)

                # add to pending commands
                pending_commands[db_num].append(command)

        # Create the threadpool and pipe jobs into it
        pool = ThreadPool(min(self._workers, len(pending_commands)))

        # execute our pending commands either in the pool, or using a pipeline
        for db_num, command_list in pending_commands.iteritems():
            if pipelined:
                pipes[db_num] = self._cluster[db_num].get_pipeline()
            for command in command_list:
                if pipelined:
                    # add to pipeline
                    pipes[db_num].add(command)
                else:
                    # execute in pool
                    pool.add(command._ident, command._execute, [self._cluster[db_num]])

        # We need to finalize our commands with a single execute in pipelines
        if pipelined:
            for db, pipe in pipes.iteritems():
                pool.add(db, pipe.execute, (), {})

        # Consolidate commands with their appropriate results
        result_map = pool.join()

        # Results get grouped by their command signature, so we have to separate the logic
        if pipelined:
            for db, result in result_map.iteritems():
                if len(result) == 1:
                    result = result[0]
                for i, value in enumerate(result):
                    command_map[pipe_command_map[db][i]]._set_value(value)

        else:
            for command in self._commands:
                result = result_map[command._ident]
                if len(result) == 1:
                    result = result[0]
                command._set_value(result)

        self._complete = True

    def get_results(self):
        assert self._complete, 'you must execute the commands before fetching results'

        return self._commands


class DistributedContextManager(object):
    def __init__(self, cluster, workers=None):
        self._workers = workers
        self._cluster = cluster

    def __enter__(self):
        self._handler = DistributedConnection(self._cluster, self._workers)
        return self._handler

    def __exit__(self, exc_type, exc_value, tb):
        # we need to break up each command and route it
        self._handler._execute()


class LazyConnectionHandler(dict):
    """
    Maps clusters of connections within a dictionary.
    """
    def __init__(self, conf_callback):
        self.conf_callback = conf_callback
        self.conf_settings = {}
        self._is_ready = False

    def __getitem__(self, key):
        if not self.is_ready():
            self.reload()
        return super(LazyConnectionHandler, self).__getitem__(key)

    def __getattr__(self, key):
        if not self.is_ready():
            self.reload()
        return super(LazyConnectionHandler, self).__getattr__(key)

    def is_ready(self):
        return self._is_ready
        # if self.conf_settings != self.conf_callback():
        #     return False
        # return True

    def reload(self):
        for conn_alias, conn_settings in self.conf_callback().iteritems():
            self[conn_alias] = create_cluster(conn_settings)
        self._is_ready = True

    def disconnect(self):
        """Disconnects all connections in cluster"""
        for connection in self.itervalues():
            connection.disconnect()



"""
nydus.conf
~~~~~~~~~~

:copyright: (c) 2011 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

try:
    VERSION = __import__('pkg_resources') \
        .get_distribution('nydus').version
except Exception, e:
    VERSION = 'unknown'

"""
Represents the default values for all Nydus settings.
"""

"""
Configuring Django's database connections, in your ``settings.py`` you would
do the following::

    DATABASES = {
        'default': {
            'ENGINE': 'nydus.contrib.django.backend',
            'NAME': 'django/default',
        },
    }

    NYDUS_CONFIG = {
        'CONNECTIONS': {
            'django/default': {
                'engine': 'nydus.contrib.django.DjangoDatabase',
                'hosts': {
                    0: {'host': 'localhost', 'backend': 'django.db.backends.sqlite3'},
                    1: {'host': 'localhost', 'backend': 'django.db.backends.sqlite3'},
                    2: {'host': 'localhost', 'backend': 'django.db.backends.sqlite3'},
                },
            },
        },
    }
"""

import warnings


CONNECTIONS = {
    # 'django/default': {
    #     'engine': 'nydus.contrib.django.backend.DjangoDatabase',
    #     'hosts': {
    #         0: {'host': 'localhost', 'backend': 'django.db.backends.sqlite3'},
    #         1: {'host': 'localhost', 'backend': 'django.db.backends.sqlite3'},
    #         2: {'host': 'localhost', 'backend': 'django.db.backends.sqlite3'},
    #     },
    # }
    # 'redis': {
    #     'engine': 'nydus.db.backends.redis.Redis',
    #     'hosts': {
    #         0: {},
    #         1: {},
    #         2: {},
    #     },
    # }
}


def configure(kwargs):
    for k, v in kwargs.iteritems():
        if k.upper() != k:
            warnings.warn('Invalid setting, \'%s\' which is not defined by Nydus' % k)
        elif k not in globals():
            warnings.warn('Invalid setting, \'%s\' which is not defined by Nydus' % k)
        else:
            globals()[k] = v



"""
nydus.db
~~~~~~~~

Disqus generic connections wrappers.

>>> from nydus.db import create_cluster
>>> redis = create_cluster({
>>>     'engine': 'nydus.db.backends.redis.Redis',
>>> })
>>> res = conn.incr('foo')
>>> assert res == 1

:copyright: (c) 2011 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

__all__ = ('create_cluster', 'connections', 'Cluster')

from nydus import conf
from .base import LazyConnectionHandler, create_cluster, Cluster

connections = LazyConnectionHandler(lambda: conf.CONNECTIONS)



from __future__ import absolute_import

from nydus.db.backends.thoonk import Thoonk
from nydus.db.base import create_cluster
import unittest2
from uuid import uuid4


class TestThoonkPubsub(unittest2.TestCase):
    """docstring for ThoonkPubsub"""
    def get_rand_name(self, prefix=None, suffix=None):
        return prefix if prefix else "" + uuid4.hex + suffix if suffix else ""

    def get_cluster(self, router):
        cluster = create_cluster({
            'engine': 'nydus.db.backends.thoonk.Thoonk',
            'router': router,
            'hosts': {
                0: {'db': 5},
                1: {'db': 6},
                2: {'db': 7},
                3: {'db': 8},
                4: {'db': 9},
            }
        })
        self.flush_custer(cluster)
        return cluster

    def flush_custer(self, cluster):
        for x in range(len(cluster)):
            if isinstance(cluster.get_conn(), list):
                c = cluster.get_conn()[x]
            else:
                c = cluster.get_conn()
            c.redis.flushdb()

    def setUp(self):
        self.ps = Thoonk(num=0, db_num=1)
        self.redis = self.ps.redis
        self.redis.flushdb()

    def tearDown(self):
        pass

    def test_job_with_ConsistentHashingRouter(self):
        pubsub = self.get_cluster('nydus.db.routers.redis.ConsistentHashingRouter')
        job = pubsub.job("test1")
        jid = job.put(10)

        jid_found = False

        for ps in pubsub.get_conn():
            jps = ps.job('test1')
            if jid in jps.get_ids():
                self.assertFalse(jid_found)
                jid_found = True

        self.assertTrue(jid_found)

    def test_job_with_RoundRobinRouter(self):
        pubsub = self.get_cluster('nydus.db.routers.redis.RoundRobinRouter')

        jobs = {}
        size = 20

        # put jobs onto the queue
        for x in xrange(0, size):
            jps = pubsub.job('testjob')
            jid = jps.put(x)
            if id(jps) not in jobs:
                jobs[id(jps)] = []
            jobs[id(jps)].append(jid)

        # make sure that we are reusing the job items
        self.assertEqual(len(jobs), 5)
        for k, v in jobs.iteritems():
            self.assertEqual(len(v), size / 5)

        # make sure we fishi
        for x in xrange(0, size):
            jps = pubsub.job('testjob')
            jid, job = jps.get()
            jps.finish(jid)

        self.assertEqual(len(jobs), 5)

        for x in range(len(pubsub)):
            ps = pubsub.get_conn()
            jps = ps.job('testjob')
            self.assertEqual(jps.get_ids(), [])


test_suite = TestThoonkPubsub



from __future__ import absolute_import

from tests import BaseTest
from nydus.db.base import Cluster
from nydus.db.backends import BaseConnection
from nydus.db.routers.redis import ConsistentHashingRouter
from nydus.db.routers.redis import RoundRobinRouter


class DummyConnection(BaseConnection):
    def __init__(self, i):
        self.host = 'dummyhost'
        self.i = i
        super(DummyConnection, self).__init__(i)

    @property
    def identifier(self):
        return "%s:%s" % (self.host, self.i)


class RoundRobinRouterTest(BaseTest):

    def setUp(self):
        self.router = RoundRobinRouter()
        self.hosts = dict((i, DummyConnection(i)) for i in range(5))
        self.cluster = Cluster(router=self.router, hosts=self.hosts)

    def get_db(self, *args, **kwargs):
        kwargs.setdefault('cluster', self.cluster)
        return self.router.get_db(*args, **kwargs)


class ConsistentHashingRouterTest(BaseTest):

    def setUp(self):
        self.router = ConsistentHashingRouter()
        self.hosts = dict((i, DummyConnection(i)) for i in range(5))
        self.cluster = Cluster(router=self.router, hosts=self.hosts)

    def get_db(self, **kwargs):
        kwargs.setdefault('cluster', self.cluster)
        return self.router.get_db(func='info', **kwargs)


class RoundRobinTest(BaseTest):

    def setUp(self):
        self.router = RoundRobinRouter()
        self.hosts = dict((i, DummyConnection(i)) for i in range(5))
        self.cluster = Cluster(router=self.router, hosts=self.hosts)

    def get_db(self, *args, **kwargs):
        kwargs.setdefault('cluster', self.cluster)
        return self.router.get_db(*args, **kwargs)

    def test_cluster_of_zero_returns_zero(self):
        self.cluster.hosts = dict()
        self.assertEquals([], self.get_db())

    def test_cluster_of_one_returns_one(self):
        self.cluster.hosts = {0: DummyConnection('foo')}
        self.assertEquals([0], self.get_db())

    def test_multi_node_cluster_returns_correct_host(self):
        self.cluster.hosts = {0: DummyConnection('foo'), 1: DummyConnection('bar')}
        self.assertEquals([[0], [1], [0], [1]], [self.get_db(), self.get_db(), self.get_db(), self.get_db(), ])


class InterfaceTest(ConsistentHashingRouterTest):

    def test_offers_router_interface(self):
        self.assertTrue(callable(self.router.get_db))

    def test_get_db_returns_itereable(self):
        iter(self.get_db())

    def test_returns_whole_cluster_without_key(self):
        self.assertEquals(range(5), self.get_db())

    def test_returns_sequence_with_one_item_when_given_key(self):
        self.assert_(len(self.get_db(key='foo')) is 1)


class HashingTest(ConsistentHashingRouterTest):

    def get_db(self, **kwargs):
        kwargs['key'] = 'foo'
        return super(HashingTest, self).get_db(**kwargs)

    def test_cluster_of_zero_returns_zero(self):
        self.cluster.hosts = dict()
        self.assertEquals([], self.get_db())

    def test_cluster_of_one_returns_one(self):
        self.cluster.hosts = dict(only_key=DummyConnection('foo'))
        self.assertEquals(['only_key'], self.get_db())

    def test_multi_node_cluster_returns_correct_host(self):
        self.assertEquals([2], self.get_db())


class RetryableTest(HashingTest):

    def test_attempt_reconnect_threshold_is_set(self):
        self.assertEqual(self.router.attempt_reconnect_threshold, 100000)

    def test_retry_gives_next_host_if_primary_is_offline(self):
        self.assertEquals([2], self.get_db())
        self.assertEquals([4], self.get_db(retry_for=2))

    def test_retry_host_change_is_sticky(self):
        self.assertEquals([2], self.get_db())
        self.assertEquals([4], self.get_db(retry_for=2))

        self.assertEquals([4], self.get_db())

    def test_adds_back_down_host_once_attempt_reconnect_threshold_is_passed(self):
        ConsistentHashingRouter.attempt_reconnect_threshold = 3

        self.assertEquals([2], self.get_db())
        self.assertEquals([4], self.get_db(retry_for=2))
        self.assertEquals([4], self.get_db())

        # Router should add host 1 back to the pool now
        self.assertEquals([2], self.get_db())

        ConsistentHashingRouter.attempt_reconnect_threshold = 100000

    def test_raises_host_list_exhaused_if_no_host_can_be_found(self):
        # Kill the first 4
        [self.get_db(retry_for=i) for i in range(4)]

        # And the 5th should raise an error
        self.assertRaises(
            ConsistentHashingRouter.HostListExhaused,
            self.get_db, **dict(retry_for=4))



from __future__ import absolute_import

from tests import BaseTest

from django.conf import settings

if not settings.configured:
    settings.configure(
        DATABASES={
            'default': {
                'ENGINE': 'nydus.contrib.django.backend',
                'NAME': 'django/sqlite3',
            },
            # 'psycopg2': {
            #     'ENGINE': 'nydus.contrib.django.backend',
            #     'NAME': 'django/psycopg2',
            # },
        },
        NYDUS_CONFIG={
            'CONNECTIONS': {
                'django/sqlite3': {
                    'engine': 'nydus.contrib.django.DjangoDatabase',
                    'hosts': {
                        0: {'backend': 'django.db.backends.sqlite3', 'name': ':memory:'},
                    },
                },
                # 'django/psycopg2': {
                #     'engine': 'nydus.contrib.django.DjangoDatabase',
                #     'hosts': {
                #         0: {'backend': 'django.db.backends.postgresql_psycopg2', 'name': 'nydus'},
                #     },
                # },
            },
        },
        # HACK: this fixes our threaded runserver remote tests
        # DATABASE_NAME='test_sentry',
        # TEST_DATABASE_NAME='test_sentry',
        INSTALLED_APPS=[
            'nydus.contrib.django',

            'tests',
        ],
        ROOT_URLCONF='',
        DEBUG=False,
        SITE_ID=1,
        TEMPLATE_DEBUG=True,
    )

from nydus.contrib.django import DjangoDatabase
from nydus.db.base import Cluster


class DjangoConnectionsTest(BaseTest):
    def test_simple(self):
        from django.db import connections

        cursor = connections['default'].execute('SELECT 1')
        self.assertEquals(cursor.fetchone(), (1,))


class DjangoSQLiteTest(BaseTest):
    def setUp(self):
        from django.db.backends import sqlite3

        self.db = DjangoDatabase(sqlite3, name=':memory:', num=0)

    def test_proxy(self):
        cursor = self.db.execute('SELECT 1')
        self.assertEquals(cursor.fetchone(), (1,))

    def test_with_cluster(self):
        p = Cluster(
            hosts={0: self.db},
        )
        cursor = p.execute('SELECT 1')
        self.assertEquals(cursor.fetchone(), (1,))

    def test_provides_identififer(self):
        self.assertEqual(
            "django.db.backends.sqlite3NAME=:memory: PORT=None HOST=None USER=None TEST_NAME=None PASSWORD=None OPTIONS={}",
            self.db.identifier
        )

# class DjangoPsycopg2Test(BaseTest):
#     def setUp(self):
#         from django.db.backends import postgresql_psycopg2
#
#         self.db = DjangoDatabase(postgresql_psycopg2, name='nydus_test', num=0)
#
#     def test_proxy(self):
#         cursor = self.db.execute('SELECT 1')
#         self.assertEquals(cursor.fetchone(), (1,))
#
#     def test_with_cluster(self):
#         p = Cluster(
#             hosts={0: self.db},
#         )
#         cursor = p.execute('SELECT 1')
#         self.assertEquals(cursor.fetchone(), (1,))



"""
nydus.db.base
~~~~~~~~~~~~~

:copyright: (c) 2011 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""


__all__ = ('BaseRouter',)


class BaseRouter(object):
    """
    Handles routing requests to a specific connection in a single cluster.
    """
    retryable = False

    def get_db(self, cluster, func, *args, **kwargs):
        """
        Return the first entry in the cluster
        The return value must be iterable
        """
        return range(len(cluster))



"""
nydus.contrib.django.backend.base
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

from __future__ import absolute_import

from django.db import DEFAULT_DB_ALIAS
from django.core.exceptions import ImproperlyConfigured

from nydus.db import connections


class DatabaseWrapper(object):
    def __init__(self, settings_dict, alias=DEFAULT_DB_ALIAS):
        self.settings_dict = settings_dict
        self.alias = alias
        try:
            nydus_alias = settings_dict['NAME']
            self.backend = connections[nydus_alias]
        except KeyError, e:
            # TODO: give a better exception
            raise ImproperlyConfigured(e)

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        return getattr(self.backend, name)



from collections import defaultdict
from Queue import Queue, Empty
from threading import Thread


# import_string comes form Werkzeug
# http://werkzeug.pocoo.org
def import_string(import_name, silent=False):
    """Imports an object based on a string. If *silent* is True the return
    value will be None if the import fails.

    Simplified version of the function with same name from `Werkzeug`_.

    :param import_name:
        The dotted name for the object to import.
    :param silent:
        If True, import errors are ignored and None is returned instead.
    :returns:
        The imported object.
    """
    import_name = str(import_name)
    try:
        if '.' in import_name:
            module, obj = import_name.rsplit('.', 1)
            return getattr(__import__(module, None, None, [obj]), obj)
        else:
            return __import__(import_name)
    except (ImportError, AttributeError):
        if not silent:
            raise


class Worker(Thread):
    def __init__(self, queue):
        Thread.__init__(self)
        self.queue = queue
        self.results = defaultdict(list)

    def run(self):
        while True:
            try:
                ident, func, args, kwargs = self.queue.get_nowait()
            except Empty:
                break

            try:
                result = func(*args, **kwargs)
                self.results[ident].append(result)
            except Exception, e:
                self.results[ident].append(e)
            finally:
                self.queue.task_done()

        return self.results


class ThreadPool(object):
    def __init__(self, workers=10):
        self.queue = Queue()
        self.workers = []
        self.tasks = []
        for worker in xrange(workers):
            self.workers.append(Worker(self.queue))

    def add(self, ident, func, args=None, kwargs=None):
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        task = (ident, func, args, kwargs)
        self.tasks.append(ident)
        self.queue.put_nowait(task)

    def join(self):
        for worker in self.workers:
            worker.start()

        results = defaultdict(list)
        for worker in self.workers:
            worker.join()
            for k, v in worker.results.iteritems():
                results[k].extend(v)
        return results



"""
nydus.contrib.django
~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

from __future__ import absolute_import

from nydus import conf
from nydus.db.backends import BaseConnection
from nydus.utils import import_string

from django.conf import settings

conf.configure(getattr(settings, 'NYDUS_CONFIG', {}))


class DjangoDatabase(BaseConnection):
    def __init__(self, backend, name, host=None, port=None, test_name=None,
                       user=None, password=None, options={}, **kwargs):
        """
        Given an alias (which is defined in DATABASES), creates a new connection
        that proxies the original database engine.
        """
        if isinstance(backend, basestring):
            backend = import_string(backend)
        self.backend = backend
        self.settings_dict = {
            'HOST': host,
            'PORT': port,
            'NAME': name,
            'TEST_NAME': test_name,
            'OPTIONS': options,
            'USER': user,
            'PASSWORD': password,
        }
        self.wrapper = __import__('%s.base' % (backend.__name__,), {}, {}, ['DatabaseWrapper']).DatabaseWrapper(self.settings_dict)
        super(DjangoDatabase, self).__init__(**kwargs)

    def connect(self):
        # force django to connect
        self.wrapper.cursor()
        return self.wrapper.connection

    def disconnect(self):
        self.connection.close()

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return self.__dict__[attr]
        else:
            # Send to router
            return getattr(self.connection, attr)

    @property
    def identifier(self):
        settings = ["%s=%s" % i for i in self.settings_dict.items()]
        return self.backend.__name__ + ' '.join(settings)



"""
nydus.db.backends.base
~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

__all__ = ('BaseConnection',)


class BasePipeline(object):
    """
    Base Pipeline class.

    This basically is absolutely useless, and just provides a sample
    API for dealing with pipelined commands.
    """
    def __init__(self, connection):
        self.pending = []
        self.connection = connection

    def add(self, command):
        self.pending.append(command)

    def execute(self):
        results = {}
        for command in self.pending:
            results[command._ident] = command(*command._args, **command._kwargs)
        return results


class BaseConnection(object):
    """
    Base connection class.

    Child classes should implement at least
    connect() and disconnect() methods.
    """

    retryable_exceptions = ()
    supports_pipelines = False

    def __init__(self, num, **options):
        self._connection = None
        self.num = num

    @property
    def identifier(self):
        return repr(self)

    @property
    def connection(self):
        if self._connection is None:
            self._connection = self.connect()
        return self._connection

    def close(self):
        if self._connection:
            self.disconnect()
        self._connection = None

    def connect(self):
        raise NotImplementedError

    def disconnect(self):
        raise NotImplementedError

    def get_pipeline(self):
        return BasePipeline(self)

    def __getattr__(self, attr):
        if self is not self.connection:
            return getattr(self.connection, attr)
        # break infinite recursion loop
        raise AttributeError("'%s' object has no attribute '%s'" % (type(self), attr))



"""
nydus.db.routers.redis
~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

from binascii import crc32
from itertools import cycle

from nydus.db.routers import BaseRouter
from nydus.contrib.ketama import Ketama


class PartitionRouter(BaseRouter):
    def get_db(self, cluster, func, key=None, *args, **kwargs):
        # Assume first argument is a key
        if not key:
            return range(len(cluster))
        return [crc32(str(key)) % len(cluster)]


class RoundRobinRouter(BaseRouter):

    def _get_db__round_robin(self, cluster):
        c = cycle(range(len(cluster)))
        for x in c:
            yield x

    def get_db(self, cluster, *args, **kwargs):
        if not cluster:
            return []
        if not hasattr(self, 'cycler'):
            self.cycler = self._get_db__round_robin(cluster)
        return [self.cycler.next()]


class ConsistentHashingRouter(BaseRouter):
    '''
    Router that returns host number based on a consistent hashing algorithm.
    The consistent hashing algorithm only works if a key argument is provided.
    If a key is not provided, then all hosts are returned.
    '''

    # Raised if all hosts in the hash have been marked as down
    class HostListExhaused(Exception):
        pass

    # If this router can be retried on if a particular db index it gave out did
    # not work
    retryable = True

    # How many requests to serve in a situation when a host is down before
    # the down hosts are retried
    attempt_reconnect_threshold = 100000

    def __init__(self):
        self._get_db_attempts = 0
        self._down_connections = set()

    # There is one instance of this class that lives inside the Cluster object
    def get_db(self, cluster, func, key=None, *args, **kwargs):
        self._setup_hash_and_connections(cluster, *args, **kwargs)

        if not cluster:
            return []
        elif not key:
            return range(len(cluster))
        else:
            return self._host_indexes_for(key, cluster)

    def flush_down_connections(self):
        for connection in self._down_connections:
            self._hash.add_node(connection.identifier)

        self._down_connections = set()

    def _setup_hash_and_connections(self, cluster, *args, **kwargs):
        # Create the hash if it doesn't exist yet
        if not hasattr(self, '_hash'):
            strings = [h.identifier for (i, h) in cluster.hosts.items()]
            self._hash = Ketama(strings)

        self._handle_host_retries(cluster, retry_for=kwargs.get('retry_for'))

    def _handle_host_retries(self, cluster, retry_for):
        self._get_db_attempts += 1

        if self._get_db_attempts > self.attempt_reconnect_threshold:
            self.flush_down_connections()
            self._get_db_attempts = 0

        if retry_for is not None:
            self._mark_connection_as_down(cluster[retry_for])

    def _mark_connection_as_down(self, connection):
        self._hash.remove_node(connection.identifier)
        self._down_connections.add(connection)

    def _host_indexes_for(self, key, cluster):
        found = self._hash.get_node(key)

        if not found and len(self._down_connections) > 0:
            raise self.HostListExhaused

        return [i for (i, h) in cluster.hosts.items()
                if h.identifier == found]



"""
nydus.db.backends.thoonk
~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

from __future__ import absolute_import

from thoonk import Pubsub

from redis import exceptions as client_exceptions

from nydus.db.backends import BaseConnection


class Thoonk(BaseConnection):
    # Exceptions that can be retried by this backend
    retryable_exceptions = client_exceptions
    supports_pipelines = False

    def __init__(self, host='localhost', port=6379, db=0, timeout=None, listen=False, **options):
        self.host = host
        self.port = port
        self.db = db
        self.timeout = timeout
        self.pubsub = None
        self.listen = listen
        super(Thoonk, self).__init__(**options)

    @property
    def identifier(self):
        mapping = vars(self)
        mapping['klass'] = self.__class__.__name__
        return "redis://%(host)s:%(port)s/%(db)s" % mapping

    def connect(self):
        return Pubsub(host=self.host, port=self.port, db=self.db, listen=self.listen)

    def disconnect(self):
        self.connection.close()

    def flushdb(self):
        """the tests assume this function exists for all redis-like backends"""
        self.connection.redis.flushdb()



"""
tests.test_connections
~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""
from __future__ import absolute_import

from mock import Mock

from nydus.db.base import Cluster, create_cluster, EventualCommand
from nydus.db.routers.base import BaseRouter
from nydus.db.backends.base import BaseConnection

from tests import BaseTest


class DummyConnection(BaseConnection):
    def __init__(self, resp='foo', **kwargs):
        self.resp = resp
        super(DummyConnection, self).__init__(**kwargs)

    def foo(self, *args, **kwargs):
        return self.resp


class DummyRouter(BaseRouter):
    def get_db(self, cluster, func, key=None, *args, **kwargs):
        # Assume first argument is a key
        if key == 'foo':
            return [1]
        return [0]


class ClusterTest(BaseTest):
    def test_create_cluster(self):
        c = create_cluster({
            'engine': DummyConnection,
            'router': DummyRouter,
            'hosts': {
                0: {'resp': 'bar'},
            }
        })
        self.assertEquals(len(c), 1)

    def test_init(self):
        c = Cluster(
            hosts={0: BaseConnection(num=1)},
        )
        self.assertEquals(len(c), 1)

    def test_proxy(self):
        c = DummyConnection(num=1, resp='bar')
        p = Cluster(
            hosts={0: c},
        )
        self.assertEquals(p.foo(), 'bar')

    def test_disconnect(self):
        c = Mock()
        p = Cluster(
            hosts={0: c},
        )
        p.disconnect()
        c.disconnect.assert_called_once()

    def test_with_router(self):
        c = DummyConnection(num=0, resp='foo')
        c2 = DummyConnection(num=1, resp='bar')

        # test dummy router
        r = DummyRouter()
        p = Cluster(
            hosts={0: c, 1: c2},
            router=r,
        )
        self.assertEquals(p.foo(), 'foo')
        self.assertEquals(p.foo('foo'), 'bar')

        # test default routing behavior
        p = Cluster(
            hosts={0: c, 1: c2},
        )
        self.assertEquals(p.foo(), ['foo', 'bar'])
        self.assertEquals(p.foo('foo'), ['foo', 'bar'])

    def test_get_conn(self):
        c = DummyConnection(alias='foo', num=0, resp='foo')
        c2 = DummyConnection(alias='foo', num=1, resp='bar')

        # test dummy router
        r = DummyRouter()
        p = Cluster(
            hosts={0: c, 1: c2},
            router=r,
        )
        self.assertEquals(p.get_conn(), c)
        self.assertEquals(p.get_conn('foo'), c2)

        # test default routing behavior
        p = Cluster(
            hosts={0: c, 1: c2},
        )
        self.assertEquals(p.get_conn(), [c, c2])
        self.assertEquals(p.get_conn('foo'), [c, c2])

    def test_map(self):
        c = DummyConnection(num=0, resp='foo')
        c2 = DummyConnection(num=1, resp='bar')

        # test dummy router
        r = DummyRouter()
        p = Cluster(
            hosts={0: c, 1: c2},
            router=r,
        )
        with p.map() as conn:
            foo = conn.foo()
            bar = conn.foo('foo')
            self.assertEquals(foo, None)
            self.assertEquals(bar, None)

        self.assertEquals(bar, 'bar')
        self.assertEquals(foo, 'foo')

        # test default routing behavior
        p = Cluster(
            hosts={0: c, 1: c2},
        )
        with p.map() as conn:
            foo = conn.foo()
            bar = conn.foo('foo')
            self.assertEquals(foo, None)
            self.assertEquals(bar, None)

        self.assertEquals(foo, ['foo', 'bar'])
        self.assertEquals(bar, ['foo', 'bar'])


class FlakeyConnection(DummyConnection):

    retryable_exceptions = [Exception]

    def foo(self, *args, **kwargs):
        if hasattr(self, 'already_failed'):
            super(FlakeyConnection, self).foo()
        else:
            self.already_failed = True
            raise Exception('boom!')


class RetryableRouter(DummyRouter):
    retryable = True

    def __init__(self):
        self.kwargs_seen = []
        self.key_args_seen = []
        super(RetryableRouter, self).__init__()

    def get_db(self, cluster, func, key=None, *args, **kwargs):
        self.kwargs_seen.append(kwargs)
        self.key_args_seen.append(key)
        return [0]


class InconsistentRouter(DummyRouter):
    retryable = True

    def __init__(self):
        self.returned = False
        super(InconsistentRouter, self).__init__()

    def get_db(self, cluster, func, key=None, *args, **kwargs):
        if self.returned:
            return range(5)
        else:
            self.returned = True
            return [0]


class ScumbagConnection(DummyConnection):

    retryable_exceptions = [Exception]

    def foo(self):
        raise Exception("Says it's a connection / Never actually connects.")


class RetryClusterTest(BaseTest):

    def build_cluster(self, connection=FlakeyConnection, router=RetryableRouter):
        return create_cluster({
            'engine': connection,
            'router': router,
            'hosts': {
                0: {'resp': 'bar'},
            }
        })

    def test_returns_correctly(self):
        cluster = self.build_cluster(connection=DummyConnection)
        self.assertEquals(cluster.foo(), 'bar')

    def test_retry_router_when_receives_error(self):
        cluster = self.build_cluster()

        cluster.foo()
        self.assertEquals({'retry_for': 0}, cluster.router.kwargs_seen.pop())

    def test_protection_from_infinate_loops(self):
        cluster = self.build_cluster(connection=ScumbagConnection)
        self.assertRaises(Exception, cluster.foo)

    def test_retryable_router_returning_multiple_dbs_raises_ecxeption(self):
        cluster = self.build_cluster(router=InconsistentRouter, connection=ScumbagConnection)
        self.assertRaisesRegexp(Exception, 'returned multiple DBs',
                                cluster.foo)


class EventualCommandTest(BaseTest):
    def test_unevaled_repr(self):
        ec = EventualCommand('foo')
        ec('bar', baz='foo')

        self.assertEquals(repr(ec), u"<EventualCommand: foo args=('bar',) kwargs={'baz': 'foo'}>")

    def test_evaled_repr(self):
        ec = EventualCommand('foo')
        ec('bar', baz='foo')
        ec._set_value('biz')

        self.assertEquals(repr(ec), u"'biz'")

    def test_coersion(self):
        ec = EventualCommand('foo')()
        ec._set_value('5')

        self.assertEquals(int(ec), 5)

    def test_nonzero(self):
        ec = EventualCommand('foo')()
        ec._set_value(None)

        self.assertEquals(int(ec or 0), 0)

    def test_evaled_unicode(self):
        ec = EventualCommand('foo')
        ec._set_value('biz')

        self.assertEquals(unicode(ec), u'biz')



"""
nydus.db.backends.redis
~~~~~~~~~~~~~~~~~~~~~~~

:copyright: (c) 2011 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

from __future__ import absolute_import

from redis import Redis as RedisClient
from redis import exceptions as client_exceptions

from nydus.db.backends import BaseConnection, BasePipeline


class RedisPipeline(BasePipeline):
    def __init__(self, connection):
        self.pending = []
        self.connection = connection
        self.pipe = connection.pipeline()

    def add(self, command):
        self.pending.append(command)
        getattr(self.pipe, command._attr)(*command._args, **command._kwargs)

    def execute(self):
        return self.pipe.execute()


class Redis(BaseConnection):
    # Exceptions that can be retried by this backend
    retryable_exceptions = client_exceptions
    supports_pipelines = True

    def __init__(self, host='localhost', port=6379, db=0, timeout=None, **options):
        self.host = host
        self.port = port
        self.db = db
        self.timeout = timeout
        super(Redis, self).__init__(**options)

    @property
    def identifier(self):
        mapping = vars(self)
        mapping['klass'] = self.__class__.__name__
        return "redis://%(host)s:%(port)s/%(db)s" % mapping

    def connect(self):
        return RedisClient(host=self.host, port=self.port, db=self.db, socket_timeout=self.timeout)

    def disconnect(self):
        self.connection.disconnect()

    def get_pipeline(self, *args, **kwargs):
        return RedisPipeline(self)



#!/usr/bin/env python

from setuptools import setup, find_packages

tests_require = [
    'nose',
    'unittest2',
    'mock',
    'redis',
    'Django>=1.2,<1.4',
    'psycopg2',
    'thoonk',
]

dependency_links = [
    'https://github.com/andyet/thoonk.py/tarball/master#egg=thoonk',
]


install_requires = [
]

setup(
    name='nydus',
    version='0.7.1',
    author='David Cramer',
    author_email='dcramer@gmail.com',
    url='http://github.com/disqus/nydus',
    description='Connection utilities',
    packages=find_packages(exclude=('tests',)),
    zip_safe=False,
    install_requires=install_requires,
    dependency_links=dependency_links,
    tests_require=tests_require,
    extras_require={'test': tests_require},
    test_suite='nose.collector',
    include_package_data=True,
    license='Apache License 2.0',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: OS Independent',
        'Topic :: Software Development'
    ],
)



from __future__ import absolute_import

from tests import BaseTest

from nydus.db.backends.redis import Redis
from nydus.db.base import Cluster, create_cluster
import mock
import redis


class RedisTest(BaseTest):

    def setUp(self):
        self.redis = Redis(num=0, db_num=1)
        self.redis.flushdb()

    def test_proxy(self):
        self.assertEquals(self.redis.incr('RedisTest_proxy'), 1)

    def test_with_cluster(self):
        p = Cluster(
            hosts={0: self.redis},
        )
        self.assertEquals(p.incr('RedisTest_with_cluster'), 1)

    def test_provides_retryable_exceptions(self):
        self.assertEquals(Redis.retryable_exceptions, redis.exceptions)

    def test_provides_identifier(self):
        self.assertEquals(self.redis.identifier, str(self.redis.identifier))

    def test_pipelined_map(self):
        redis = create_cluster({
            'engine': 'nydus.db.backends.redis.Redis',
            'router': 'nydus.db.routers.redis.PartitionRouter',
            'hosts': {
                0: {'db': 5},
                1: {'db': 6},
                2: {'db': 7},
                3: {'db': 8},
                4: {'db': 9},
            }
        })
        chars = ('a', 'b', 'c', 'd', 'e', 'f')
        with redis.map() as conn:
            [conn.set(c, i) for i, c in enumerate(chars)]
            res = [conn.get(c) for c in chars]
        self.assertEqual(range(len(chars)), [int(r._wrapped) for r in res])

    @mock.patch('nydus.db.backends.redis.RedisClient')
    def test_client_instantiates_with_kwargs(self, RedisClient):
        client = Redis(num=0)
        client.connect()

        self.assertEquals(RedisClient.call_count, 1)
        RedisClient.assert_any_call(host='localhost', port=6379, db=0, socket_timeout=None)

    @mock.patch('nydus.db.backends.redis.RedisClient')
    def test_map_does_pipeline(self, RedisClient):
        redis = create_cluster({
            'engine': 'nydus.db.backends.redis.Redis',
            'router': 'nydus.db.routers.redis.PartitionRouter',
            'hosts': {
                0: {'db': 0},
                1: {'db': 1},
            }
        })
        with redis.map() as conn:
            conn.set('a', 0)
            conn.set('d', 1)

        # ensure this was actually called through the pipeline
        self.assertFalse(RedisClient().set.called)

        self.assertEquals(RedisClient().pipeline.call_count, 2)
        RedisClient().pipeline.assert_called_with()

        self.assertEquals(RedisClient().pipeline().set.call_count, 2)
        RedisClient().pipeline().set.assert_any_call('a', 0)
        RedisClient().pipeline().set.assert_any_call('d', 1)

        self.assertEquals(RedisClient().pipeline().execute.call_count, 2)
        RedisClient().pipeline().execute.assert_called_with()

    @mock.patch('nydus.db.backends.redis.RedisClient')
    def test_map_only_runs_on_required_nodes(self, RedisClient):
        redis = create_cluster({
            'engine': 'nydus.db.backends.redis.Redis',
            'router': 'nydus.db.routers.redis.PartitionRouter',
            'hosts': {
                0: {'db': 0},
                1: {'db': 1},
            }
        })
        with redis.map() as conn:
            conn.set('a', 0)
            conn.set('b', 1)

        # ensure this was actually called through the pipeline
        self.assertFalse(RedisClient().set.called)

        self.assertEquals(RedisClient().pipeline.call_count, 1)
        RedisClient().pipeline.assert_called_with()

        self.assertEquals(RedisClient().pipeline().set.call_count, 2)
        RedisClient().pipeline().set.assert_any_call('a', 0)
        RedisClient().pipeline().set.assert_any_call('b', 1)

        self.assertEquals(RedisClient().pipeline().execute.call_count, 1)
        RedisClient().pipeline().execute.assert_called_with()



from __future__ import absolute_import

from nydus.db.backends.base import BaseConnection
from mock import patch
from mock import Mock

import unittest2


class TestBaseConnection(unittest2.TestCase):
    def test_getattr_break_recursion(self):

        with patch.object(BaseConnection, 'connection') as mock_connection:
            b = BaseConnection(0)
            mock_connection.__get__ = Mock(return_value=b)

            with self.assertRaises(AttributeError):
                print b.foo



"""
nydus
~~~~~

:copyright: (c) 2011 DISQUS.
:license: Apache License 2.0, see LICENSE for more details.
"""

try:
    VERSION = __import__('pkg_resources') \
        .get_distribution('nydus').version
except Exception, e:
    VERSION = 'unknown'



"""
    Ketama consistent hash algorithm.

    Rewrited from the original source: http://www.audioscrobbler.net/development/ketama/

"""
__author__  = "Andrey Nikishaev"
__email__   = "creotiv@gmail.com"
__version__ = 0.1
__status__  = "productrion"

__all__     = ['Ketama']

import hashlib
import math
from bisect import bisect

class Ketama(object):

    def __init__(self, nodes=None, weights=None):
        """
            nodes   - List of nodes(strings)
            weights - Dictionary of node wheights where keys are nodes names.
                      if not set, all nodes will be equal.
        """
        self._hashring      = dict()
        self._sorted_keys   = []

        self._nodes         = nodes
        self._weights       = weights if weights else {}

        self._build_circle()

    def _build_circle(self):
        """
            Creates hash ring.
        """
        total_weight = 0
        for node in self._nodes:
            total_weight += self._weights.get(node, 1)

        for node in self._nodes:
            weight = self._weights.get(node,1)

            ks = math.floor((40*len(self._nodes) * weight) / total_weight);

            for i in xrange(0, int(ks)):
                b_key = self._md5_digest( '%s-%s-salt' % (node, i) )

                for l in xrange(0, 4):
                    key = (( b_key[3+l*4] << 24 )
                         | ( b_key[2+l*4] << 16 )
                         | ( b_key[1+l*4] << 8  )
                         |   b_key[l*4]         )

                    self._hashring[key] = node
                    self._sorted_keys.append(key)

        self._sorted_keys.sort()

    def _get_node_pos(self, key):
        """
            Return node position(integer) for a given key. Else return None
        """
        if not self._hashring:
            return None

        key = self._gen_key(key)

        nodes = self._sorted_keys
        pos = bisect(nodes, key)

        if pos == len(nodes):
            return 0
        return pos

    def _gen_key(self, key):
        """
            Return long integer for a given key, that represent it place on
            the hash ring.
        """
        b_key = self._md5_digest(key)
        return self._hashi(b_key, lambda x: x)

    def _hashi(self, b_key, fn):
        return (( b_key[fn(3)] << 24 )
              | ( b_key[fn(2)] << 16 )
              | ( b_key[fn(1)] << 8  )
              |   b_key[fn(0)]       )

    def _md5_digest(self, key):
        return map(ord, hashlib.md5(key).digest())

    def remove_node(self,node):
        """
            Removes node from circle and rebuild it.
        """
        try:
            self._nodes.remove(node)
            del self._weights[node]
        except KeyError,e:
            pass
        except ValueError,e:
            pass
        self._hashring    = dict()
        self._sorted_keys = []

        self._build_circle()

    def add_node(self,node,weight=1):
        """
            Adds node to circle and rebuild it.
        """
        self._nodes.append(node)
        self._weights[node] = weight
        self._hashring      = dict()
        self._sorted_keys   = []

        self._build_circle()

    def get_node(self, key):
        """
            Return node for a given key. Else return None.
        """
        pos = self._get_node_pos(key)
        if pos is None:
            return None
        return self._hashring[ self._sorted_keys[pos] ]


if __name__ == '__main__':
    def test(k):
        data = {}
        for i in xrange(REQUESTS):
            tower = k.get_node('a'+str(i))
            data.setdefault(tower,0)
            data[tower] += 1
        print 'Number of caches on each node: '
        print data
        print ''

        print k.get_node('Aplple');
        print k.get_node('Hello');
        print k.get_node('Data');
        print k.get_node('Computer');

    NODES = ['192.168.0.1:6000','192.168.0.1:6001','192.168.0.1:6002',
            '192.168.0.1:6003','192.168.0.1:6004','192.168.0.1:6005',
            '192.168.0.1:6006','192.168.0.1:6008','192.168.0.1:6007'
           ]
    REQUESTS = 1000

    k = Ketama(NODES)

    test(k)

    k.remove_node('192.168.0.1:6007')

    test(k)

    k.add_node('192.168.0.1:6007')

    test(k)
